# coding:utf-8

__project__ = "xkits-file"
__version__ = "0.2"
__urlhome__ = "https://github.com/bondbox/xfiles/"
__description__ = "File module"

# author
__author__ = "Mingzhe Zou"
__author_email__ = "zoumingzhe@outlook.com"
